@@ -1,4 +1,6 @@
 from tnetwork.readwrite.SN_graph_io import *
 from tnetwork.readwrite.IG_graph_io import *
 from tnetwork.readwrite.SN_com_io import *
-from tnetwork.readwrite.IG_com_io import *
\ No newline at end of file
+from tnetwork.readwrite.IG_com_io import *
+from tnetwork.readwrite.generic_io import *
+from tnetwork.readwrite.LS_graph_io import *
\ No newline at end of file
